@@ -1,12 +1,20 @@
 import torch
 from torch import nn
 
-def expand_dim(t, dim, k):
-    t = t.unsqueeze(dim)
+def expand_dim(t, dim, k, unsqueeze = False):
+    if unsqueeze:
+        t = t.unsqueeze(dim)
     expand_shape = [-1] * len(t.shape)
     expand_shape[dim] = k
     return t.expand(*expand_shape)
 
+def fetch_pkm_value_parameters(module):
+    params = []
+    for m in module.modules():
+        if isinstance(m, PKM):
+            params.append(m.values.weight)
+    return params
+
 class PKM(nn.Module):
     def __init__(self, dim, heads = 8, num_keys = 128, topk = 10):
         super().__init__()
@@ -16,22 +24,25 @@ class PKM(nn.Module):
         self.num_keys = num_keys
 
         d_head = dim // heads
-        self.to_queries = nn.Linear(dim, dim * 2, bias = False)
-        self.to_out = nn.Linear(dim, dim)
+        self.batch_norm = nn.BatchNorm1d(dim)
+        self.to_queries = nn.Linear(dim, dim, bias = False)
 
-        self.keys = nn.Parameter(torch.randn(heads, num_keys, 2, d_head))
-        self.values = nn.Parameter(torch.randn(heads, num_keys ** 2, d_head))
+        self.keys = nn.Parameter(torch.randn(heads, num_keys, 2, d_head // 2))
+        self.values = nn.EmbeddingBag(num_keys ** 2, dim, mode='sum')
 
     def forward(self, x):
         b, t, e, h = *x.shape, self.heads
-        queries = self.to_queries(x).chunk(2, dim=-1)
+        queries = self.to_queries(x)
+        queries = self.batch_norm(queries.transpose(1, 2)).transpose(1, 2)
+        queries = queries.chunk(2, dim=-1)
         queries = torch.stack(queries).reshape(2, b, t, h, -1)
 
-        dots = torch.einsum('pbthd,hnpd->bhtpn', queries, self.keys)
+        dots = torch.einsum('pbthd,hnpd->bthpn', queries, self.keys)
         scores, indices = dots.topk(k=self.topk, dim=-1)
         scores, indices = map(lambda x: x.chunk(2, dim=2), (scores, indices))
 
-        shape = (b, h, t, self.topk ** 2)
+        all_topk = self.topk ** 2
+        shape = (b, t, h, all_topk)
 
         all_scores = (
             scores[0][..., :, None] +
@@ -48,10 +59,7 @@ class PKM(nn.Module):
 
         attn = final_topk.softmax(dim=-1)
 
-        values = self.values[None, :, None, :, :].expand(b, -1, t, -1, -1)
-        expanded_indices = expand_dim(value_indices, dim=4, k=values.shape[-1])
-        selected_values = values.gather(-2, expanded_indices)
+        value_indices, attn = map(lambda x: x.reshape(-1, self.topk * h), (value_indices, attn))
+        out = self.values(value_indices, per_sample_weights=attn)
+        return out.reshape(b, t, e)
 
-        out = (attn.unsqueeze(-1) * selected_values).sum(dim=-2)
-        out = out.transpose(1, 2).reshape(b, t, -1)
-        return self.to_out(out)
